/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.viewmodel;

import jalview.api.AlignViewportI;
import jalview.datamodel.AlignmentI;

/**
 * Slightly less embryonic class which: Supplies and updates viewport properties
 * relating to position such as: start and end residues and sequences; ideally
 * will serve hidden columns/rows too. Intention also to support calculations
 * for positioning, scrolling etc. such as finding the middle of the viewport,
 * checking for scrolls off screen
 */
public class ViewportRanges extends ViewportProperties
{
  // start residue of viewport
  private int startRes;

  // end residue of viewport
  private int endRes;

  // start sequence of viewport
  private int startSeq;

  // end sequence of viewport
  private int endSeq;

  // alignment
  private AlignmentI al;

  /**
   * Constructor
   * 
   * @param alignment
   *          the viewport's alignment
   */
  public ViewportRanges(AlignmentI alignment)
  {
    // initial values of viewport settings
    this.startRes = 0;
    this.endRes = alignment.getWidth() - 1;
    this.startSeq = 0;
    this.endSeq = alignment.getHeight() - 1;
    this.al = alignment;
  }

  /**
   * Get alignment width in cols, including hidden cols
   */
  public int getAbsoluteAlignmentWidth()
  {
    return al.getWidth();
  }

  /**
   * Get alignment height in rows, including hidden rows
   */
  public int getAbsoluteAlignmentHeight()
  {
    return al.getHeight() + al.getHiddenSequences().getSize();
  }

  /**
   * Set first residue visible in the viewport, and retain the current width.
   * 
   * @param res
   *          residue position
   */
  public void setStartRes(int res)
  {
    int width = getViewportWidth();
    setStartEndRes(res, res + width - 1);
  }

  /**
   * Set start and end residues at the same time. This method only fires one
   * event for the two changes, and should be used in preference to separate
   * calls to setStartRes and setEndRes.
   * 
   * @param start
   *          the start residue
   * @param end
   *          the end residue
   */
  public void setStartEndRes(int start, int end)
  {
    int oldres = this.startRes;
    if (start > al.getWidth() - 1)
    {
      startRes = al.getWidth() - 1;
    }
    else if (start < 0)
    {
      startRes = 0;
    }
    else
    {
      startRes = start;
    }

    if (end < 0)
    {
      endRes = 0;
    }
    else
    {
      endRes = end;
    }

    changeSupport.firePropertyChange("startres", oldres, start);
  }

  /**
   * Set last residue visible in the viewport
   * 
   * @param res
   *          residue position
   */
  public void setEndRes(int res)
  {
    int width = getViewportWidth();
    setStartEndRes(res - width + 1, res);
  }

  /**
   * Set the first sequence visible in the viewport
   * 
   * @param seq
   *          sequence position
   */
  public void setStartSeq(int seq)
  {
    int height = getViewportHeight();
    setStartEndSeq(seq, seq + height - 1);
  }

  /**
   * Set start and end sequences at the same time. This method only fires one
   * event for the two changes, and should be used in preference to separate
   * calls to setStartSeq and setEndSeq.
   * 
   * @param start
   *          the start sequence
   * @param end
   *          the end sequence
   */
  public void setStartEndSeq(int start, int end)
  {
    int oldseq = this.startSeq;
    if (start > al.getHeight() - 1)
    {
      startSeq = al.getHeight() - 1;
    }
    else if (start < 0)
    {
      startSeq = 0;
    }
    else
    {
      startSeq = start;
    }

    if (end >= al.getHeight())
    {
      endSeq = al.getHeight() - 1;
    }
    else if (end < 0)
    {
      endSeq = 0;
    }
    else
    {
      endSeq = end;
    }

    changeSupport.firePropertyChange("startseq", oldseq, start);
  }

  /**
   * Set the last sequence visible in the viewport
   * 
   * @param seq
   *          sequence position
   */
  public void setEndSeq(int seq)
  {
    int height = getViewportHeight();
    setStartEndSeq(seq - height + 1, seq);
  }

  /**
   * Get start residue of viewport
   */
  public int getStartRes()
  {
    return startRes;
  }

  /**
   * Get end residue of viewport
   */
  public int getEndRes()
  {
    return endRes;
  }

  /**
   * Get start sequence of viewport
   */
  public int getStartSeq()
  {
    return startSeq;
  }

  /**
   * Get end sequence of viewport
   */
  public int getEndSeq()
  {
    return endSeq;
  }

  /**
   * Set viewport width in residues, without changing startRes. Use in
   * preference to calculating endRes from the width, to avoid out by one
   * errors!
   * 
   * @param w
   *          width in residues
   */
  public void setViewportWidth(int w)
  {
    setStartEndRes(startRes, startRes + w - 1);
  }

  /**
   * Set viewport height in residues, without changing startSeq. Use in
   * preference to calculating endSeq from the height, to avoid out by one
   * errors!
   * 
   * @param h
   *          height in sequences
   */
  public void setViewportHeight(int h)
  {
    setStartEndSeq(startSeq, startSeq + h - 1);
  }

  /**
   * Set viewport horizontal start position and width. Use in preference to
   * calculating endRes from the width, to avoid out by one errors!
   * 
   * @param start
   *          start residue
   * @param w
   *          width in residues
   */
  public void setViewportStartAndWidth(int start, int w)
  {
    setStartEndRes(start, start + w - 1);
  }

  /**
   * Set viewport vertical start position and height. Use in preference to
   * calculating endSeq from the height, to avoid out by one errors!
   * 
   * @param start
   *          start sequence
   * @param h
   *          height in sequences
   */
  public void setViewportStartAndHeight(int start, int h)
  {
    setStartEndSeq(start, start + h - 1);
  }

  /**
   * Get width of viewport in residues
   * 
   * @return width of viewport
   */
  public int getViewportWidth()
  {
    return (endRes - startRes + 1);
  }

  /**
   * Get height of viewport in residues
   * 
   * @return height of viewport
   */
  public int getViewportHeight()
  {
    return (endSeq - startSeq + 1);
  }

  /**
   * Scroll the viewport range vertically
   * 
   * @param up
   *          true if scrolling up, false if down
   * 
   * @return true if the scroll is valid
   */
  public boolean scrollUp(boolean up)
  {
    if (up)
    {
      if (startSeq < 1)
      {
        return false;
      }

      setStartSeq(startSeq - 1);
    }
    else
    {
      if (endSeq >= al.getHeight() - 1)
      {
        return false;
      }

      setStartSeq(startSeq + 1);
    }
    return true;
  }

  /**
   * Scroll the viewport range horizontally
   * 
   * @param right
   *          true if scrolling right, false if left
   * 
   * @return true if the scroll is valid
   */
  public boolean scrollRight(boolean right)
  {
    if (!right)
    {
      if (startRes < 1)
      {
        return false;
      }

      setStartRes(startRes - 1);
    }
    else
    {
      if (endRes > al.getWidth() - 1)
      {
        return false;
      }

      setStartRes(startRes + 1);
    }

    return true;
  }

  /**
   * Scroll a wrapped alignment so that the specified residue is visible
   * 
   * @param res
   *          residue position to scroll to
   */
  public void scrollToWrappedVisible(int res)
  {
    // get the start residue of the wrapped row which res is in
    // and set that as our start residue
    int width = getViewportWidth();
    setStartRes((res / width) * width);
  }

  // pass av here until hidden columns JAL-2388 merged, then use alignment
  // instead
  public void scrollToVisible(int x, int y, AlignViewportI av)
  {
    while (y < startSeq)
    {
      scrollUp(true);
    }
    while (y + 1 > endSeq)
    {
      scrollUp(false);
    }

    while (x < av.getColumnSelection().adjustForHiddenColumns(startRes))
    {
      if (!scrollRight(false))
      {
        break;
      }
    }
    while (x > av.getColumnSelection().adjustForHiddenColumns(endRes))
    {
      if (!scrollRight(true))
      {
        break;
      }
    }
  }

}
